/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*- */
/* GdkPixbuf library - animated avi support
 *
 * Copyright (C) 1999 The Free Software Foundation
 *
 * Authors: Jonathan Blandford <jrb@redhat.com>
 *          Havoc Pennington <hp@redhat.com>
 *
 * Slightly edited io-gif-animation.c by Jrgen Fogh
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <config.h>
#include <errno.h>
#include "gdk-pixbuf-private.h"
#include "io-avi-animation.h"

static void gdk_pixbuf_avi_anim_class_init (GdkPixbufAviAnimClass *klass);
static void gdk_pixbuf_avi_anim_finalize   (GObject        *object);

static gboolean                gdk_pixbuf_avi_anim_is_static_image  (GdkPixbufAnimation *animation);
static GdkPixbuf*              gdk_pixbuf_avi_anim_get_static_image (GdkPixbufAnimation *animation);

static void                    gdk_pixbuf_avi_anim_get_size (GdkPixbufAnimation *anim,
                                                             int                *width,
                                                             int                *height);
static GdkPixbufAnimationIter* gdk_pixbuf_avi_anim_get_iter (GdkPixbufAnimation *anim,
                                                             const GTimeVal     *start_time);




static gpointer parent_class;

GType
gdk_pixbuf_avi_anim_get_type (void)
{
        static GType object_type = 0;

        if (!object_type) {
                static const GTypeInfo object_info = {
                        sizeof (GdkPixbufAviAnimClass),
                        (GBaseInitFunc) NULL,
                        (GBaseFinalizeFunc) NULL,
                        (GClassInitFunc) gdk_pixbuf_avi_anim_class_init,
                        NULL,           /* class_finalize */
                        NULL,           /* class_data */
                        sizeof (GdkPixbufAviAnim),
                        0,              /* n_preallocs */
                        (GInstanceInitFunc) NULL,
                };
                
                object_type = g_type_register_static (GDK_TYPE_PIXBUF_ANIMATION,
                                                      "GdkPixbufAviAnim",
                                                      &object_info, 0);
        }
        
        return object_type;
}

static void
gdk_pixbuf_avi_anim_class_init (GdkPixbufAviAnimClass *klass)
{
        GObjectClass *object_class = G_OBJECT_CLASS (klass);
        GdkPixbufAnimationClass *anim_class = GDK_PIXBUF_ANIMATION_CLASS (klass);
        
        parent_class = g_type_class_peek_parent (klass);
        
        object_class->finalize = gdk_pixbuf_avi_anim_finalize;

        anim_class->is_static_image = gdk_pixbuf_avi_anim_is_static_image;
        anim_class->get_static_image = gdk_pixbuf_avi_anim_get_static_image;
        anim_class->get_size = gdk_pixbuf_avi_anim_get_size;
        anim_class->get_iter = gdk_pixbuf_avi_anim_get_iter;
}

static void
gdk_pixbuf_avi_anim_finalize (GObject *object)
{
        GdkPixbufAviAnim *avi_anim = GDK_PIXBUF_AVI_ANIM (object);

        GList *l;
        GdkPixbufFrame *frame;
        
        for (l = avi_anim->frames; l; l = l->next) {
                frame = l->data;
                g_object_unref (frame->pixbuf);
                g_free (frame);
        }
        
        g_list_free (avi_anim->frames);
        
        G_OBJECT_CLASS (parent_class)->finalize (object);
}

static gboolean
gdk_pixbuf_avi_anim_is_static_image  (GdkPixbufAnimation *animation)
{
        GdkPixbufAviAnim *avi_anim;

        avi_anim = GDK_PIXBUF_AVI_ANIM (animation);

        return (avi_anim->frames != NULL &&
                avi_anim->frames->next == NULL);
}

static GdkPixbuf*
gdk_pixbuf_avi_anim_get_static_image (GdkPixbufAnimation *animation)
{
        GdkPixbufAviAnim *avi_anim;

        avi_anim = GDK_PIXBUF_AVI_ANIM (animation);

        if (avi_anim->frames == NULL)
                return NULL;
        else
                return GDK_PIXBUF (((GdkPixbufFrame*)avi_anim->frames->data)->pixbuf);        
}

static void
gdk_pixbuf_avi_anim_get_size (GdkPixbufAnimation *anim,
                              int                *width,
                              int                *height)
{
        GdkPixbufAviAnim *avi_anim;

        avi_anim = GDK_PIXBUF_AVI_ANIM (anim);

        if (width)
                *width = avi_anim->width;

        if (height)
                *height = avi_anim->height;
}


static void
iter_clear (GdkPixbufAviAnimIter *iter)
{
        iter->current_frame = NULL;
}

static void
iter_restart (GdkPixbufAviAnimIter *iter)
{
        iter_clear (iter);
  
        iter->current_frame = iter->avi_anim->frames;
}

static GdkPixbufAnimationIter*
gdk_pixbuf_avi_anim_get_iter (GdkPixbufAnimation *anim,
                              const GTimeVal     *start_time)
{
        GdkPixbufAviAnimIter *iter;

        iter = g_object_new (GDK_TYPE_PIXBUF_AVI_ANIM_ITER, NULL);

        iter->avi_anim = GDK_PIXBUF_AVI_ANIM (anim);

        g_object_ref (iter->avi_anim);
        
        iter_restart (iter);

        iter->start_time = *start_time;
        iter->current_time = *start_time;
        iter->first_loop_slowness = 0;
        
        return GDK_PIXBUF_ANIMATION_ITER (iter);
}



static void gdk_pixbuf_avi_anim_iter_class_init (GdkPixbufAviAnimIterClass *klass);
static void gdk_pixbuf_avi_anim_iter_finalize   (GObject                   *object);

static int        gdk_pixbuf_avi_anim_iter_get_delay_time             (GdkPixbufAnimationIter *iter);
static GdkPixbuf* gdk_pixbuf_avi_anim_iter_get_pixbuf                 (GdkPixbufAnimationIter *iter);
static gboolean   gdk_pixbuf_avi_anim_iter_on_currently_loading_frame (GdkPixbufAnimationIter *iter);
static gboolean   gdk_pixbuf_avi_anim_iter_advance                    (GdkPixbufAnimationIter *iter,
                                                                       const GTimeVal         *current_time);



static gpointer iter_parent_class;

GType
gdk_pixbuf_avi_anim_iter_get_type (void)
{
        static GType object_type = 0;

        if (!object_type) {
                static const GTypeInfo object_info = {
                        sizeof (GdkPixbufAviAnimIterClass),
                        (GBaseInitFunc) NULL,
                        (GBaseFinalizeFunc) NULL,
                        (GClassInitFunc) gdk_pixbuf_avi_anim_iter_class_init,
                        NULL,           /* class_finalize */
                        NULL,           /* class_data */
                        sizeof (GdkPixbufAviAnimIter),
                        0,              /* n_preallocs */
                        (GInstanceInitFunc) NULL,
                };
                
                object_type = g_type_register_static (GDK_TYPE_PIXBUF_ANIMATION_ITER,
                                                      "GdkPixbufAviAnimIter",
                                                      &object_info, 0);
        }
        
        return object_type;
}

static void
gdk_pixbuf_avi_anim_iter_class_init (GdkPixbufAviAnimIterClass *klass)
{
        GObjectClass *object_class = G_OBJECT_CLASS (klass);
        GdkPixbufAnimationIterClass *anim_iter_class =
                GDK_PIXBUF_ANIMATION_ITER_CLASS (klass);
        
        iter_parent_class = g_type_class_peek_parent (klass);
        
        object_class->finalize = gdk_pixbuf_avi_anim_iter_finalize;

        anim_iter_class->get_delay_time = gdk_pixbuf_avi_anim_iter_get_delay_time;
        anim_iter_class->get_pixbuf = gdk_pixbuf_avi_anim_iter_get_pixbuf;
        anim_iter_class->on_currently_loading_frame = gdk_pixbuf_avi_anim_iter_on_currently_loading_frame;
        anim_iter_class->advance = gdk_pixbuf_avi_anim_iter_advance;
}

static void
gdk_pixbuf_avi_anim_iter_finalize (GObject *object)
{
        GdkPixbufAviAnimIter *iter = GDK_PIXBUF_AVI_ANIM_ITER (object);

        iter_clear (iter);

        g_object_unref (iter->avi_anim);
        
        G_OBJECT_CLASS (iter_parent_class)->finalize (object);
}

static gboolean
gdk_pixbuf_avi_anim_iter_advance (GdkPixbufAnimationIter *anim_iter,
                                  const GTimeVal         *current_time)
{
        GdkPixbufAviAnimIter *iter;
        gint elapsed;
        gint loop;
        GList *tmp;
        GList *old;
        
        iter = GDK_PIXBUF_AVI_ANIM_ITER (anim_iter);
        
        iter->current_time = *current_time;

        /* We use milliseconds for all times */
        elapsed =
          (((iter->current_time.tv_sec - iter->start_time.tv_sec) * G_USEC_PER_SEC +
            iter->current_time.tv_usec - iter->start_time.tv_usec)) / 1000;

        if (elapsed < 0) {
                /* Try to compensate; probably the system clock
                 * was set backwards
                 */
                iter->start_time = iter->current_time;
                elapsed = 0;
        }

        g_assert (iter->avi_anim->total_time > 0);
        
        /* See how many times we've already played the full animation,
         * and subtract time for that.
         */

        /* If current_frame is NULL at this point, we have loaded the
         * animation from a source which fell behind the speed of the 
         * display. We remember how much slower the first loop was due
         * to this and correct the position calculation in order to not
         * jump in the middle of the second loop.
         */
        if (iter->current_frame == NULL)
                iter->first_loop_slowness = MAX(0, elapsed - iter->avi_anim->total_time);

        loop = (elapsed - iter->first_loop_slowness) / iter->avi_anim->total_time;
        elapsed = (elapsed - iter->first_loop_slowness) % iter->avi_anim->total_time;

        iter->position = elapsed;

        /* Now move to the proper frame */
        if (iter->avi_anim->loop == 0 || loop < iter->avi_anim->loop) 
                tmp = iter->avi_anim->frames;
        else 
                tmp = NULL;
        while (tmp != NULL) {
                GdkPixbufFrame *frame = tmp->data;
                
                if (iter->position >= frame->elapsed &&
                    iter->position < (frame->elapsed + frame->delay_time))
                        break;
                
                tmp = tmp->next;
        }

        old = iter->current_frame;
        
        iter->current_frame = tmp;

        return iter->current_frame != old;
}

int
gdk_pixbuf_avi_anim_iter_get_delay_time (GdkPixbufAnimationIter *anim_iter)
{
        GdkPixbufFrame *frame;
        GdkPixbufAviAnimIter *iter;
  
        iter = GDK_PIXBUF_AVI_ANIM_ITER (anim_iter);

        if (iter->current_frame) {
                frame = iter->current_frame->data;

#if 0
                g_print ("frame start: %d pos: %d frame len: %d frame remaining: %d\n",
                         frame->elapsed,
                         iter->position,
                         frame->delay_time,
                         frame->delay_time - (iter->position - frame->elapsed));
#endif
                
                return frame->delay_time - (iter->position - frame->elapsed);
        } else 
                return -1; /* show last frame forever */
}

GdkPixbuf*
gdk_pixbuf_avi_anim_iter_get_pixbuf (GdkPixbufAnimationIter *anim_iter)
{
        GdkPixbufAviAnimIter *iter;
        GdkPixbufFrame *frame;
        
        iter = GDK_PIXBUF_AVI_ANIM_ITER (anim_iter);

        frame = iter->current_frame ? iter->current_frame->data : g_list_last (iter->avi_anim->frames)->data;
/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*- */
/* GdkPixbuf library - AVI loader declarations
 *
 * Copyright (C) 1999 The Free Software Foundation
 *
 * Authors: Mark Crichton <crichton@gimp.org>
 *          Miguel de Icaza <miguel@gnu.org>
 *          Federico Mena-Quintero <federico@gimp.org>
 *          Havoc Pennington <hp@redhat.com>
 *
 * Slightly edited io-gif-animation.h by Jrgen Fogh
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef GDK_PIXBUF_AVI_H
#define GDK_PIXBUF_AVI_H

#include "gdk-pixbuf-animation.h"

typedef struct _GdkPixbufAviAnim GdkPixbufAviAnim;
typedef struct _GdkPixbufAviAnimClass GdkPixbufAviAnimClass;
typedef struct _GdkPixbufFrame GdkPixbufFrame;

#define GDK_TYPE_PIXBUF_AVI_ANIM              (gdk_pixbuf_avi_anim_get_type ())
#define GDK_PIXBUF_AVI_ANIM(object)           (G_TYPE_CHECK_INSTANCE_CAST ((object), GDK_TYPE_PIXBUF_AVI_ANIM, GdkPixbufAviAnim))
#define GDK_IS_PIXBUF_AVI_ANIM(object)        (G_TYPE_CHECK_INSTANCE_TYPE ((object), GDK_TYPE_PIXBUF_AVI_ANIM))

#define GDK_PIXBUF_AVI_ANIM_CLASS(klass)      (G_TYPE_CHECK_CLASS_CAST ((klass), GDK_TYPE_PIXBUF_AVI_ANIM, GdkPixbufAviAnimClass))
#define GDK_IS_PIXBUF_AVI_ANIM_CLASS(klass)   (G_TYPE_CHECK_CLASS_TYPE ((klass), GDK_TYPE_PIXBUF_AVI_ANIM))
#define GDK_PIXBUF_AVI_ANIM_GET_CLASS(obj)    (G_TYPE_INSTANCE_GET_CLASS ((obj), GDK_TYPE_PIXBUF_AVI_ANIM, GdkPixbufAviAnimClass))

/* Private part of the GdkPixbufAviAnim structure */
struct _GdkPixbufAviAnim {
        GdkPixbufAnimation parent_instance;

        /* Number of frames */
        int n_frames;

        /* Total length of animation */
        int total_time;
        
	/* List of GdkPixbufFrame structures */
        GList *frames;

	/* bounding box size */
	int width, height;
        
        int loop;
};

struct _GdkPixbufAviAnimClass {
        GdkPixbufAnimationClass parent_class;
        
};

GType gdk_pixbuf_avi_anim_get_type (void) G_GNUC_CONST;



typedef struct _GdkPixbufAviAnimIter GdkPixbufAviAnimIter;
typedef struct _GdkPixbufAviAnimIterClass GdkPixbufAviAnimIterClass;


#define GDK_TYPE_PIXBUF_AVI_ANIM_ITER              (gdk_pixbuf_avi_anim_iter_get_type ())
#define GDK_PIXBUF_AVI_ANIM_ITER(object)           (G_TYPE_CHECK_INSTANCE_CAST ((object), GDK_TYPE_PIXBUF_AVI_ANIM_ITER, GdkPixbufAviAnimIter))
#define GDK_IS_PIXBUF_AVI_ANIM_ITER(object)        (G_TYPE_CHECK_INSTANCE_TYPE ((object), GDK_TYPE_PIXBUF_AVI_ANIM_ITER))

#define GDK_PIXBUF_AVI_ANIM_ITER_CLASS(klass)      (G_TYPE_CHECK_CLASS_CAST ((klass), GDK_TYPE_PIXBUF_AVI_ANIM_ITER, GdkPixbufAviAnimIterClass))
#define GDK_IS_PIXBUF_AVI_ANIM_ITER_CLASS(klass)   (G_TYPE_CHECK_CLASS_TYPE ((klass), GDK_TYPE_PIXBUF_AVI_ANIM_ITER))
#define GDK_PIXBUF_AVI_ANIM_ITER_GET_CLASS(obj)    (G_TYPE_INSTANCE_GET_CLASS ((obj), GDK_TYPE_PIXBUF_AVI_ANIM_ITER, GdkPixbufAviAnimIterClass))

struct _GdkPixbufAviAnimIter {
        GdkPixbufAnimationIter parent_instance;
        
        GdkPixbufAviAnim   *avi_anim;

        GTimeVal            start_time;
        GTimeVal            current_time;

        /* Time in milliseconds into this run of the animation */
        gint                position;
        
        GList              *current_frame;
        
        gint                first_loop_slowness;
};

struct _GdkPixbufAviAnimIterClass {
        GdkPixbufAnimationIterClass parent_class;

};

GType gdk_pixbuf_avi_anim_iter_get_type (void) G_GNUC_CONST;



struct _GdkPixbufFrame {
	/* The pixbuf with this frame's image data */
	GdkPixbuf *pixbuf;

	/* Frame duration in ms */
	int delay_time;

        /* Sum of preceding delay times */
        int elapsed;
};

#endif

#if 0
        if (FALSE && frame)
          g_print ("current frame %d dispose mode %d  %d x %d\n",
                   g_list_index (iter->avi_anim->frames,
                                 frame),
                   frame->action,
                   gdk_pixbuf_get_width (frame->pixbuf),
                   gdk_pixbuf_get_height (frame->pixbuf));
#endif
        
        if (frame == NULL)
                return NULL;
        
        return frame->pixbuf;
}

static gboolean
gdk_pixbuf_avi_anim_iter_on_currently_loading_frame (GdkPixbufAnimationIter *anim_iter)
{
        GdkPixbufAviAnimIter *iter;
  
        iter = GDK_PIXBUF_AVI_ANIM_ITER (anim_iter);

        return iter->current_frame == NULL || iter->current_frame->next == NULL;  
}
