/* GdkPixbuf library - Windows Bitmap image loader
 *
 * Copyright (C) 1999 Jrgen Fogh
 *
 * Authors: Jrgen Fogh <jfogh@mail.dk>
 *
 * Based on io-bmp.c and io-ani.c
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#undef DEBUG_AVI

#include <string.h>
#include "gdk-pixbuf-io.h"
#include "gdk-pixbuf-private.h"
#include "io-avi-animation.h"

static int
lsb_32 (guchar *src)
{
	return src[0] | (src[1] << 8) | (src[2] << 16) | (src[3] << 24);
}

#define MAKE_TAG(a,b,c,d) ( (guint32)d << 24 | \
			    (guint32)c << 16 | \
			    (guint32)b <<  8 | \
			    (guint32)a )

#define TAG_RIFF MAKE_TAG('R','I','F','F')
#define TAG_AVI  MAKE_TAG('A','V','I',' ')
#define TAG_hdrl MAKE_TAG('h','d','r','l')
#define TAG_LIST MAKE_TAG('L','I','S','T')
#define TAG_movi MAKE_TAG('m','o','v','i')
#define TAG_avih MAKE_TAG('a','v','i','h')
#define TAG_strl MAKE_TAG('s','t','r','l')
#define TAG_strh MAKE_TAG('s','t','r','h')
#define TAG_strd MAKE_TAG('s','t','r','d')
#define TAG_strn MAKE_TAG('s','t','r','n')
#define TAG_vids MAKE_TAG('v','i','d','s')
#define TAG_DIB  MAKE_TAG('D','I','B',' ')
#define TAG_strf MAKE_TAG('s','t','r','f')
#define TAG_JUNK MAKE_TAG('J','U','N','K')
#define TAG_00db MAKE_TAG('0','0','d','b')
#define TAG_idx1 MAKE_TAG('i','d','x','1')

typedef struct {
	guint32 Type;
	guint32 Handler;
	guint16 Language;
	guint32 InitialFrames;
	guint32 Scale;
	guint32 Rate; /* dwRate / dwScale == samples/second */
	guint32 Start;
	guint32 Length; /* In units above... */
	guint32 SuggestedBufferSize;
	guint32 Quality;
	guint32 SampleSize;
	guint32 Frame;
} StreamHeader;


typedef struct _AviLoaderContext AviLoaderContext;
struct _AviLoaderContext {
        guint32 data_size;
        guint32 NumFrames; 

	StreamHeader stream_header;
	GdkPixbuf *pixbuf;
	GdkPixbufAviAnim *animation;
	guint elapsed;
	guint delay_time;
	guint width;
	guint height;
	guchar depth;
	guint Lines;		/* # of finished lines */
        guint32 cp;

        guint32 chunk_id;
        guint32 chunk_size;

        guchar *buffer;
        guchar *byte;
        guint   n_bytes;
        guint   buffer_size;

	GdkPixbufModulePreparedFunc prepared_func;
	GdkPixbufModuleUpdatedFunc updated_func;
	gpointer user_data;
};

static gpointer gdk_pixbuf__avi_image_begin_load (GdkPixbufModuleSizeFunc size_func,
					   GdkPixbufModulePreparedFunc prepared_func,
					   GdkPixbufModuleUpdatedFunc updated_func,
					   gpointer user_data,
					   GError **error);
static gboolean
gdk_pixbuf__avi_image_stop_load (gpointer data,
				 GError **error);

static gboolean gdk_pixbuf__avi_image_load_increment (gpointer data,
						      const guchar *buf, guint size,
						      GError **error);

#define BYTES_LEFT(context) \
        ((context)->n_bytes - ((context)->byte - (context)->buffer))

static void
read_int8 (AviLoaderContext *context,
           guchar     *data,
           int        count)
{
        int total = MIN (count, BYTES_LEFT (context));
        memcpy (data, context->byte, total);
        context->byte += total;
        context->cp += total;
}

static guint32
read_int32 (AviLoaderContext *context)
{
        guint32 result;

        read_int8 (context, (guchar*) &result, 4);
        return lsb_32 ((guchar *) &result);
}

static void
context_free (AviLoaderContext *context)
{
        if (!context)
                return;

        if (context->animation) 
		g_object_unref (context->animation);
        g_free (context->buffer);
        
        g_free (context);
}

static void
OneLine24 (AviLoaderContext *context)
{
	gint X;
	guchar *Pixels;

	X = 0;
	Pixels = (context->pixbuf->pixels +
		  context->pixbuf->rowstride *
		  (context->height - context->Lines - 1));
	while (X < context->width) {
		Pixels[X * 3 + 0] = context->byte[X * 3 + 2];
		Pixels[X * 3 + 1] = context->byte[X * 3 + 1];
		Pixels[X * 3 + 2] = context->byte[X * 3 + 0];
		X++;
	}
	context->byte += context->chunk_size / context->height;
	context->cp += context->chunk_size / context->height;
}

static void
OneLine16 (AviLoaderContext *context)
{
	int i;
	guchar *pixels;
	guchar *src;

	pixels = (context->pixbuf->pixels +
		  context->pixbuf->rowstride * (context->height - context->Lines - 1));

	src = context->byte;

	for (i = 0; i < context->width; i++) {
		int v, r, g, b;

		v = src[0] | (src[1] << 8);

		r = (v >> 10) & 0x1f;
		g = (v >> 5) & 0x1f;
		b = v & 0x1f;

		*pixels++ = (r << 3) | (r >> 2);
		*pixels++ = (g << 3) | (g >> 2);
		*pixels++ = (b << 3) | (b >> 2);

		src += 2;
	}
	context->byte += context->chunk_size / context->height;
	context->cp += context->chunk_size / context->height;
}

static void
prepared_notify (GdkPixbuf *pixbuf, 
		 GdkPixbufAnimation *anim, 
		 gpointer user_data)
{
	if (anim != NULL)
		g_object_ref (anim);
	*((GdkPixbufAnimation **)user_data) = anim;
}

static GdkPixbufAnimation *
gdk_pixbuf__avi_image_load_animation (FILE *f, GError **error)
{
	guchar buffer[4096];
	size_t length;
	GdkPixbufAnimation *anim = NULL;
	gpointer context;

	context = gdk_pixbuf__avi_image_begin_load (NULL, prepared_notify, 
						    NULL, &anim, error);

	if (!context)
                return NULL;
	
	while (!feof (f)) {
		length = fread (buffer, 1, sizeof (buffer), f);
		if (length > 0)
			if (!gdk_pixbuf__avi_image_load_increment (context, buffer, length, error)) {
				gdk_pixbuf__avi_image_stop_load (context, NULL);
				if (anim != NULL)
					g_object_unref (anim);
				return NULL;
			}
	}

	if (!gdk_pixbuf__avi_image_stop_load (context, error)) {
		if (anim != NULL)
			g_object_unref (anim);
		return NULL;
	}
	
	return anim;
}

static gboolean
avi_load_chunk (AviLoaderContext *context, GError **error)
{
        if (context->chunk_id == 0x0) {
                if (BYTES_LEFT (context) < 8)
                        return FALSE;
                context->chunk_id = read_int32 (context);
                context->chunk_size = read_int32 (context);
		/* Pad it up to word length */
		if (context->chunk_size % 2)
			context->chunk_size += (2 - (context->chunk_size % 2));
        
        }
        
        while (context->chunk_id == TAG_LIST)
	{
		if (BYTES_LEFT (context) < 12)
			return FALSE;
                        
		read_int32 (context);
		context->chunk_id = read_int32 (context);
		context->chunk_size = read_int32 (context);
		/* Pad it up to word length */
		if (context->chunk_size % 2)
			context->chunk_size += (2 - (context->chunk_size % 2));
	}

	if (context->chunk_id == TAG_movi) {
	  /* FIXME: Set some flag in context, so we know that the header has been read. */

	  return TRUE;
	}

        if (BYTES_LEFT (context) < context->chunk_size)
                return FALSE;

	if (context->chunk_id == TAG_strh) {
		context->stream_header.Type = read_int32 (context);
		context->stream_header.Handler = read_int32 (context);
		read_int32 (context);
		context->stream_header.Language = read_int32 (context);
		context->stream_header.InitialFrames = read_int32 (context);
		context->stream_header.Scale = read_int32 (context);
		context->stream_header.Rate = read_int32 (context);
		context->stream_header.Start = read_int32 (context);
		context->stream_header.Length = read_int32 (context);
		context->stream_header.SuggestedBufferSize = read_int32 (context);
		context->stream_header.Quality = read_int32 (context);
		context->stream_header.SampleSize = read_int32 (context);

#ifdef DEBUG_AVI
		g_print ("Type \t%" G_GUINT32_FORMAT
			 "\nHandler \t%" G_GUINT32_FORMAT
			 "\nLanguage \t%" G_GUINT32_FORMAT
			 "\nInitialFrames \t%" G_GUINT32_FORMAT
			 "\nScale \t%" G_GUINT32_FORMAT
			 "\nRate \t%" G_GUINT32_FORMAT
			 "\nStart \t%" G_GUINT32_FORMAT
			 "\nLength \t%" G_GUINT32_FORMAT
			 "\nSuggestedBufferSize \t%" G_GUINT32_FORMAT
			 "\nQuality \t%" G_GUINT32_FORMAT
			 "\nSampleSize \t%" G_GUINT32_FORMAT
			 "\n",
			 context->stream_header.Type,
			 context->stream_header.Handler,
			 context->stream_header.Language,
			 context->stream_header.InitialFrames,
			 context->stream_header.Scale,
			 context->stream_header.Rate,
			 context->stream_header.Start,
			 context->stream_header.Length,
			 context->stream_header.SuggestedBufferSize,
			 context->stream_header.Quality,
			 context->stream_header.SampleSize);
#endif

		context->stream_header.Frame = read_int32 (context);
		read_int32 (context);
		context->delay_time = 1000 * context->stream_header.Scale / context->stream_header.Rate;
		context->NumFrames = context->stream_header.Length;
		/* FIXME: Check for unsupported types. */
		if (context->NumFrames == 0 || 
		    context->NumFrames >= 1024) 
		{
			g_set_error (error,
				     GDK_PIXBUF_ERROR,
				     GDK_PIXBUF_ERROR_CORRUPT_IMAGE,
				     _("Invalid header in animation"));
			return FALSE;
		}
	}

	if (context->chunk_id == TAG_strf) {
		context->cp += 6;
		context->byte += 4;
		context->width = read_int32 (context);
		context->height = read_int32 (context);
		context->byte += 2;
		read_int8 (context, &context->depth, 1);

		context->animation->width = context->width;
		context->animation->height = context->height;

		context->chunk_size -= 15;
	}

	if (context->chunk_id == TAG_avih ||
	    context->chunk_id == TAG_strf ||
	    context->chunk_id == TAG_strd ||
	    context->chunk_id == TAG_strn ||
	    context->chunk_id == TAG_JUNK ||
	    context->chunk_id == TAG_idx1) {
		context->byte += context->chunk_size;
		context->cp += context->chunk_size;
	}

	if (context->chunk_id == TAG_00db) {
		GdkPixbufFrame *frame;

		context->pixbuf = gdk_pixbuf_new (GDK_COLORSPACE_RGB, FALSE,
						  8, context->width,
						  context->height);
                if (!context->pixbuf) {
                        g_set_error (error, GDK_PIXBUF_ERROR,
                                     GDK_PIXBUF_ERROR_INSUFFICIENT_MEMORY,
                                     _("Not enough memory to load AVI file"));
                        return -2;
                }

		for (context->Lines = 0; context->Lines < context->height; context->Lines++) {
			if (context->depth == 24)
				OneLine24 (context);
			else
				OneLine16 (context);
		}

                frame = g_new (GdkPixbufFrame, 1);

		frame->elapsed = context->elapsed;
		frame->delay_time = context->delay_time;
		context->elapsed += context->delay_time;
		frame->pixbuf = context->pixbuf;
		context->pixbuf = NULL;

		context->animation->n_frames++;
		context->animation->total_time += context->delay_time;

		if (context->animation->frames == NULL) {
			if (context->prepared_func)
				(* context->prepared_func) (frame->pixbuf, 
							    GDK_PIXBUF_ANIMATION (context->animation), 
							    context->user_data);
		}

		context->animation->frames = g_list_append (context->animation->frames, frame);
	}

	context->chunk_id = 0x0;
	return BYTES_LEFT (context) > 0;
}

static gboolean
gdk_pixbuf__avi_image_load_increment (gpointer data,
				      const guchar *buf, guint size,
				      GError **error)
{
	AviLoaderContext *context = (AviLoaderContext *)data;

        if (context->n_bytes + size >= context->buffer_size) {
                int drop = context->byte - context->buffer;
                memmove (context->buffer, context->byte, context->n_bytes - drop);
                context->n_bytes -= drop;
                context->byte = context->buffer;
                if (context->n_bytes + size >= context->buffer_size) {
			guchar *tmp;
                        context->buffer_size = MAX (context->n_bytes + size, context->buffer_size + 4096);
#ifdef DEBUG_AVI
                        g_print ("growing buffer to %" G_GUINT32_FORMAT "\n", context->buffer_size);
#endif
                        tmp = g_try_realloc (context->buffer, context->buffer_size);
                        if (!tmp) {
				g_set_error (error,
					     GDK_PIXBUF_ERROR,
					     GDK_PIXBUF_ERROR_INSUFFICIENT_MEMORY,
					     _("Not enough memory to load animation"));
				return FALSE;
			}
                        context->byte = context->buffer = tmp;
                }
        }
        memcpy (context->buffer + context->n_bytes, buf, size);
        context->n_bytes += size;

        if (context->data_size == 0) {
		guint32 riff_id, chunk_id;
                        
		if (BYTES_LEFT (context) < 12)
			return TRUE;
                        
		riff_id = read_int32 (context);
		context->data_size = read_int32 (context);
		chunk_id = read_int32 (context);
                        
		if (riff_id != TAG_RIFF || 
		    context->data_size == 0 || 
		    chunk_id != TAG_AVI) {
			g_set_error (error,
				     GDK_PIXBUF_ERROR,
				     GDK_PIXBUF_ERROR_CORRUPT_IMAGE,
				     _("Invalid header in animation"));
			return FALSE; 
		}
	}
        
        if (context->cp < context->data_size + 8) {
		GError *chunk_error = NULL;

		while (avi_load_chunk (context, &chunk_error)) ;
		if (chunk_error) {
			g_propagate_error (error, chunk_error);
			return FALSE;
		}
	}

        return TRUE;
}

static gpointer
gdk_pixbuf__avi_image_begin_load (GdkPixbufModuleSizeFunc size_func, 
                                  GdkPixbufModulePreparedFunc prepared_func, 
				  GdkPixbufModuleUpdatedFunc  updated_func,
				  gpointer user_data,
				  GError **error)
{
        AviLoaderContext *context;

        context = g_new0 (AviLoaderContext, 1);
        
        context->animation = g_object_new (GDK_TYPE_PIXBUF_AVI_ANIM, NULL);

        context->prepared_func = prepared_func;
        context->updated_func = updated_func;
        context->user_data = user_data;
        
        context->buffer_size = 4096;
        context->buffer = g_try_malloc (context->buffer_size);
        if (!context->buffer) {
		context_free (context);
		g_set_error (error,
			     GDK_PIXBUF_ERROR,
			     GDK_PIXBUF_ERROR_INSUFFICIENT_MEMORY,
			     _("Not enough memory to load animation"));
		return NULL;
	}
        
        context->byte = context->buffer;
        context->n_bytes = 0;
        
        return (gpointer) context;
}

static gboolean
gdk_pixbuf__avi_image_stop_load (gpointer data,
				 GError **error)
{
	AviLoaderContext *context = (AviLoaderContext *)data;

	context_free (context);
}

void
MODULE_ENTRY (avi, fill_vtable) (GdkPixbufModule *module)
{
/*         module->load = gdk_pixbuf__avi_image_load; */
        module->begin_load = gdk_pixbuf__avi_image_begin_load;
        module->stop_load = gdk_pixbuf__avi_image_stop_load;
        module->load_increment = gdk_pixbuf__avi_image_load_increment;
/*         module->begin_load = NULL; */
/*         module->stop_load = NULL; */
/*         module->load_increment = NULL; */
        module->load_animation = gdk_pixbuf__avi_image_load_animation;
}

void
MODULE_ENTRY (avi, fill_info) (GdkPixbufFormat *info)
{
        static GdkPixbufModulePattern signature[] = {
                { "RIFF    AVI ", "    xxxx    ", 100 },
                { NULL, NULL, 0 }
        };
	static gchar * mime_types[] = {
		"video/avi",
		NULL
	};
	static gchar * extensions[] = {
		"avi",
		NULL
	};

	info->name = "avi";
        info->signature = signature;
	info->description = N_("The AVI video format");
	info->mime_types = mime_types;
	info->extensions = extensions;
	info->flags = 0;
}
